@@ -1,11 +1,13 @@
 use std::{
     ffi::CString,
+    os::unix::io::{AsFd, BorrowedFd, OwnedFd},
     os::unix::{io::RawFd, net::UnixStream},
     sync::Arc,
 };
 
 use crate::{
     core_interfaces::{WL_CALLBACK_INTERFACE, WL_DISPLAY_INTERFACE, WL_REGISTRY_INTERFACE},
+    debug,
     protocol::{
         check_for_signature, same_interface, same_interface_or_anonymous, AllowNull, Argument,
         ArgumentType, Interface, Message, ObjectInfo, ProtocolError, ANONYMOUS_INTERFACE,
@@ -15,7 +17,6 @@ use crate::{
     types::server::{DisconnectReason, InvalidId},
 };
 
-use io_lifetimes::OwnedFd;
 use smallvec::SmallVec;
 
 use crate::rs::{
@@ -135,11 +136,12 @@ impl<D> Client<D> {
         }
 
         if self.debug {
-            crate::rs::debug::print_send_message(
+            debug::print_send_message(
                 object.interface.name,
                 object_id.id.id,
                 message_desc.name,
                 &args,
+                false,
             );
         }
 
@@ -295,13 +297,10 @@ impl<D> Client<D> {
 
     #[cfg(any(target_os = "linux", target_os = "android"))]
     pub(crate) fn get_credentials(&self) -> Credentials {
-        use std::os::unix::io::AsRawFd;
-        let creds = nix::sys::socket::getsockopt(
-            self.socket.as_raw_fd(),
-            nix::sys::socket::sockopt::PeerCredentials,
-        )
-        .expect("getsockopt failed!?");
-        Credentials { pid: creds.pid(), uid: creds.uid(), gid: creds.gid() }
+        let creds =
+            rustix::net::sockopt::get_socket_peercred(&self.socket).expect("getsockopt failed!?");
+        let pid = rustix::process::Pid::as_raw(Some(creds.pid));
+        Credentials { pid, uid: creds.uid.as_raw(), gid: creds.gid.as_raw() }
     }
 
     #[cfg(not(any(target_os = "linux", target_os = "android")))]
@@ -336,7 +335,7 @@ impl<D> Client<D> {
         &mut self,
     ) -> std::io::Result<(Message<u32, OwnedFd>, Object<Data<D>>)> {
         if self.killed {
-            return Err(nix::errno::Errno::EPIPE.into());
+            return Err(rustix::io::Errno::PIPE.into());
         }
         loop {
             let map = &self.map;
@@ -358,11 +357,21 @@ impl<D> Client<D> {
                 }
                 Err(MessageParseError::Malformed) => {
                     self.kill(DisconnectReason::ConnectionClosed);
-                    return Err(nix::errno::Errno::EPROTO.into());
+                    return Err(rustix::io::Errno::PROTO.into());
                 }
             };
 
             let obj = self.map.find(msg.sender_id).unwrap();
+
+            if self.debug {
+                debug::print_dispatched_message(
+                    obj.interface.name,
+                    msg.sender_id,
+                    obj.interface.requests.get(msg.opcode as usize).unwrap().name,
+                    &msg.args,
+                );
+            }
+
             return Ok((msg, obj));
         }
     }
@@ -649,6 +658,12 @@ impl<D> Client<D> {
     }
 }
 
+impl<D> AsFd for Client<D> {
+    fn as_fd(&self) -> BorrowedFd<'_> {
+        self.socket.as_fd()
+    }
+}
+
 #[derive(Debug)]
 pub(crate) struct ClientStore<D: 'static> {
     clients: Vec<Option<Client<D>>>,
