@@ -1,11 +1,16 @@
 //! Wayland socket manipulation
 
+use std::collections::VecDeque;
 use std::io::{ErrorKind, IoSlice, IoSliceMut, Result as IoResult};
-use std::os::unix::io::{AsRawFd, RawFd};
+use std::os::unix::io::{AsFd, AsRawFd, BorrowedFd, OwnedFd, RawFd};
 use std::os::unix::net::UnixStream;
+use std::slice;
 
-use io_lifetimes::{AsFd, BorrowedFd, OwnedFd};
-use nix::sys::socket;
+use rustix::io::retry_on_intr;
+use rustix::net::{
+    recvmsg, send, sendmsg, RecvAncillaryBuffer, RecvAncillaryMessage, RecvFlags,
+    SendAncillaryBuffer, SendAncillaryMessage, SendFlags,
+};
 
 use crate::protocol::{ArgumentType, Message};
 
@@ -34,15 +39,22 @@ impl Socket {
     /// The `fds` slice should not be longer than `MAX_FDS_OUT`, and the `bytes`
     /// slice should not be longer than `MAX_BYTES_OUT` otherwise the receiving
     /// end may lose some data.
-    pub fn send_msg(&self, bytes: &[u8], fds: &[RawFd]) -> IoResult<usize> {
-        let flags = socket::MsgFlags::MSG_DONTWAIT | socket::MsgFlags::MSG_NOSIGNAL;
-        let iov = [IoSlice::new(bytes)];
+    pub fn send_msg(&self, bytes: &[u8], fds: &[OwnedFd]) -> IoResult<usize> {
+        #[cfg(not(target_os = "macos"))]
+        let flags = SendFlags::DONTWAIT | SendFlags::NOSIGNAL;
+        #[cfg(target_os = "macos")]
+        let flags = SendFlags::DONTWAIT;
 
         if !fds.is_empty() {
-            let cmsgs = [socket::ControlMessage::ScmRights(fds)];
-            Ok(socket::sendmsg::<()>(self.stream.as_raw_fd(), &iov, &cmsgs, flags, None)?)
+            let iov = [IoSlice::new(bytes)];
+            let mut cmsg_space = vec![0; rustix::cmsg_space!(ScmRights(fds.len()))];
+            let mut cmsg_buffer = SendAncillaryBuffer::new(&mut cmsg_space);
+            let fds =
+                unsafe { slice::from_raw_parts(fds.as_ptr() as *const BorrowedFd, fds.len()) };
+            cmsg_buffer.push(SendAncillaryMessage::ScmRights(fds));
+            Ok(retry_on_intr(|| sendmsg(self, &iov, &mut cmsg_buffer, flags))?)
         } else {
-            Ok(socket::sendmsg::<()>(self.stream.as_raw_fd(), &iov, &[], flags, None)?)
+            Ok(retry_on_intr(|| send(self, bytes, flags))?)
         }
     }
 
@@ -57,33 +69,40 @@ impl Socket {
     /// The `buffer` slice should be at least `MAX_BYTES_OUT` long and the `fds`
     /// slice `MAX_FDS_OUT` long, otherwise some data of the received message may
     /// be lost.
-    pub fn rcv_msg(&self, buffer: &mut [u8], fds: &mut [RawFd]) -> IoResult<(usize, usize)> {
-        let mut cmsg = nix::cmsg_space!([RawFd; MAX_FDS_OUT]);
-        let mut iov = [IoSliceMut::new(buffer)];
-        let msg = socket::recvmsg::<()>(
-            self.stream.as_raw_fd(),
-            &mut iov[..],
-            Some(&mut cmsg),
-            socket::MsgFlags::MSG_DONTWAIT
-                | socket::MsgFlags::MSG_CMSG_CLOEXEC
-                | socket::MsgFlags::MSG_NOSIGNAL,
-        )?;
+    pub fn rcv_msg(&self, buffer: &mut [u8], fds: &mut VecDeque<OwnedFd>) -> IoResult<usize> {
+        #[cfg(not(target_os = "macos"))]
+        let flags = RecvFlags::DONTWAIT | RecvFlags::CMSG_CLOEXEC;
+        #[cfg(target_os = "macos")]
+        let flags = RecvFlags::DONTWAIT;
 
-        let mut fd_count = 0;
-        let received_fds = msg.cmsgs().flat_map(|cmsg| match cmsg {
-            socket::ControlMessageOwned::ScmRights(s) => s,
-            _ => Vec::new(),
-        });
-        for (fd, place) in received_fds.zip(fds.iter_mut()) {
-            fd_count += 1;
-            *place = fd;
+        let mut cmsg_space = vec![0; rustix::cmsg_space!(ScmRights(MAX_FDS_OUT))];
+        let mut cmsg_buffer = RecvAncillaryBuffer::new(&mut cmsg_space);
+        let mut iov = [IoSliceMut::new(buffer)];
+        let msg = retry_on_intr(|| recvmsg(&self.stream, &mut iov[..], &mut cmsg_buffer, flags))?;
+
+        let received_fds = cmsg_buffer
+            .drain()
+            .filter_map(|cmsg| match cmsg {
+                RecvAncillaryMessage::ScmRights(fds) => Some(fds),
+                _ => None,
+            })
+            .flatten();
+        fds.extend(received_fds);
+        #[cfg(target_os = "macos")]
+        for fd in fds.iter() {
+            if let Ok(flags) = rustix::io::fcntl_getfd(fd) {
+                let _ = rustix::io::fcntl_setfd(fd, flags | rustix::io::FdFlags::CLOEXEC);
+            }
         }
-        Ok((msg.bytes, fd_count))
+        Ok(msg.bytes)
     }
 }
 
 impl From<UnixStream> for Socket {
     fn from(stream: UnixStream) -> Self {
+        // macOS doesn't have MSG_NOSIGNAL, but has SO_NOSIGPIPE instead
+        #[cfg(target_os = "macos")]
+        let _ = rustix::net::sockopt::set_socket_nosigpipe(&stream, true);
         Self { stream }
     }
 }
@@ -109,10 +128,10 @@ impl AsRawFd for Socket {
 #[derive(Debug)]
 pub struct BufferedSocket {
     socket: Socket,
-    in_data: Buffer<u32>,
-    in_fds: Buffer<RawFd>,
-    out_data: Buffer<u32>,
-    out_fds: Buffer<RawFd>,
+    in_data: Buffer<u8>,
+    in_fds: VecDeque<OwnedFd>,
+    out_data: Buffer<u8>,
+    out_fds: Vec<OwnedFd>,
 }
 
 impl BufferedSocket {
@@ -120,32 +139,23 @@ impl BufferedSocket {
     pub fn new(socket: Socket) -> Self {
         Self {
             socket,
-            in_data: Buffer::new(2 * MAX_BYTES_OUT / 4), // Incoming buffers are twice as big in order to be
-            in_fds: Buffer::new(2 * MAX_FDS_OUT),        // able to store leftover data if needed
-            out_data: Buffer::new(MAX_BYTES_OUT / 4),
-            out_fds: Buffer::new(MAX_FDS_OUT),
+            in_data: Buffer::new(2 * MAX_BYTES_OUT), // Incoming buffers are twice as big in order to be
+            in_fds: VecDeque::new(),                 // able to store leftover data if needed
+            out_data: Buffer::new(MAX_BYTES_OUT),
+            out_fds: Vec::new(),
         }
     }
 
     /// Flush the contents of the outgoing buffer into the socket
     pub fn flush(&mut self) -> IoResult<()> {
         let written = {
-            let words = self.out_data.get_contents();
-            if words.is_empty() {
+            let bytes = self.out_data.get_contents();
+            if bytes.is_empty() {
                 return Ok(());
             }
-            let bytes = unsafe {
-                ::std::slice::from_raw_parts(words.as_ptr() as *const u8, words.len() * 4)
-            };
-            let fds = self.out_fds.get_contents();
-            let written = self.socket.send_msg(bytes, fds)?;
-            for &fd in fds {
-                // once the fds are sent, we can close them
-                let _ = ::nix::unistd::close(fd);
-            }
-            written
+            self.socket.send_msg(bytes, &self.out_fds)?
         };
-        self.out_data.offset(written / 4);
+        self.out_data.offset(written);
         self.out_data.move_to_front();
         self.out_fds.clear();
         Ok(())
@@ -159,14 +169,9 @@ impl BufferedSocket {
     // if false is returned, it means there is not enough space
     // in the buffer
     fn attempt_write_message(&mut self, msg: &Message<u32, RawFd>) -> IoResult<bool> {
-        match write_to_buffers(
-            msg,
-            self.out_data.get_writable_storage(),
-            self.out_fds.get_writable_storage(),
-        ) {
-            Ok((bytes_out, fds_out)) => {
+        match write_to_buffers(msg, self.out_data.get_writable_storage(), &mut self.out_fds) {
+            Ok(bytes_out) => {
                 self.out_data.advance(bytes_out);
-                self.out_fds.advance(fds_out);
                 Ok(true)
             }
             Err(MessageWriteError::BufferTooSmall) => Ok(false),
@@ -192,7 +197,7 @@ impl BufferedSocket {
             if !self.attempt_write_message(msg)? {
                 // If this fails again, this means the message is too big
                 // to be transmitted at all
-                return Err(::nix::errno::Errno::E2BIG.into());
+                return Err(rustix::io::Errno::TOOBIG.into());
             }
         }
         Ok(())
@@ -203,23 +208,17 @@ impl BufferedSocket {
     pub fn fill_incoming_buffers(&mut self) -> IoResult<()> {
         // reorganize the buffers
         self.in_data.move_to_front();
-        self.in_fds.move_to_front();
         // receive a message
-        let (in_bytes, in_fds) = {
-            let words = self.in_data.get_writable_storage();
-            let bytes = unsafe {
-                ::std::slice::from_raw_parts_mut(words.as_ptr() as *mut u8, words.len() * 4)
-            };
-            let fds = self.in_fds.get_writable_storage();
-            self.socket.rcv_msg(bytes, fds)?
+        let in_bytes = {
+            let bytes = self.in_data.get_writable_storage();
+            self.socket.rcv_msg(bytes, &mut self.in_fds)?
         };
         if in_bytes == 0 {
             // the other end of the socket was closed
-            return Err(::nix::errno::Errno::EPIPE.into());
+            return Err(rustix::io::Errno::PIPE.into());
         }
         // advance the storage
-        self.in_data.advance(in_bytes / 4 + usize::from(in_bytes % 4 > 0));
-        self.in_fds.advance(in_fds);
+        self.in_data.advance(in_bytes);
         Ok(())
     }
 
@@ -235,19 +234,17 @@ impl BufferedSocket {
     where
         F: FnMut(u32, u16) -> Option<&'static [ArgumentType]>,
     {
-        let (msg, read_data, read_fd) = {
+        let (msg, read_data) = {
             let data = self.in_data.get_contents();
-            let fds = self.in_fds.get_contents();
-            if data.len() < 2 {
+            if data.len() < 2 * 4 {
                 return Err(MessageParseError::MissingData);
             }
-            let object_id = data[0];
-            let opcode = (data[1] & 0x0000_FFFF) as u16;
+            let object_id = u32::from_ne_bytes([data[0], data[1], data[2], data[3]]);
+            let word_2 = u32::from_ne_bytes([data[4], data[5], data[6], data[7]]);
+            let opcode = (word_2 & 0x0000_FFFF) as u16;
             if let Some(sig) = signature(object_id, opcode) {
-                match parse_message(data, sig, fds) {
-                    Ok((msg, rest_data, rest_fds)) => {
-                        (msg, data.len() - rest_data.len(), fds.len() - rest_fds.len())
-                    }
+                match parse_message(data, sig, &mut self.in_fds) {
+                    Ok((msg, rest_data)) => (msg, data.len() - rest_data.len()),
                     Err(e) => return Err(e),
                 }
             } else {
@@ -257,7 +254,6 @@ impl BufferedSocket {
         };
 
         self.in_data.offset(read_data);
-        self.in_fds.offset(read_fd);
 
         Ok(msg)
     }
@@ -304,6 +300,7 @@ impl<T: Copy + Default> Buffer<T> {
     ///
     /// This only sets the counter of occupied space back to zero,
     /// allowing previous content to be overwritten.
+    #[allow(unused)]
     fn clear(&mut self) {
         self.occupied = 0;
         self.offset = 0;
@@ -323,13 +320,7 @@ impl<T: Copy + Default> Buffer<T> {
     /// maximal write space availability
     fn move_to_front(&mut self) {
         if self.occupied > self.offset {
-            unsafe {
-                ::std::ptr::copy(
-                    &self.storage[self.offset] as *const T,
-                    &mut self.storage[0] as *mut T,
-                    self.occupied - self.offset,
-                );
-            }
+            self.storage.copy_within((self.offset)..(self.occupied), 0)
         }
         self.occupied -= self.offset;
         self.offset = 0;
@@ -342,14 +333,14 @@ mod tests {
     use crate::protocol::{AllowNull, Argument, ArgumentType, Message};
 
     use std::ffi::CString;
-    use std::os::unix::io::RawFd;
+    use std::os::unix::io::BorrowedFd;
     use std::os::unix::prelude::IntoRawFd;
 
     use smallvec::smallvec;
 
-    fn same_file(a: RawFd, b: RawFd) -> bool {
-        let stat1 = ::nix::sys::stat::fstat(a).unwrap();
-        let stat2 = ::nix::sys::stat::fstat(b).unwrap();
+    fn same_file(a: BorrowedFd, b: BorrowedFd) -> bool {
+        let stat1 = rustix::fs::fstat(a).unwrap();
+        let stat2 = rustix::fs::fstat(b).unwrap();
         stat1.st_dev == stat2.st_dev && stat1.st_ino == stat2.st_ino
     }
 
@@ -366,7 +357,9 @@ mod tests {
         assert_eq!(msg1.args.len(), msg2.args.len());
         for (arg1, arg2) in msg1.args.iter().zip(msg2.args.iter()) {
             if let (Argument::Fd(fd1), Argument::Fd(fd2)) = (arg1, arg2) {
-                assert!(same_file(fd1.as_raw_fd(), fd2.as_raw_fd()));
+                let fd1 = unsafe { BorrowedFd::borrow_raw(fd1.as_raw_fd()) };
+                let fd2 = unsafe { BorrowedFd::borrow_raw(fd2.as_raw_fd()) };
+                assert!(same_file(fd1, fd2));
             } else {
                 assert_eq!(arg1, arg2);
             }
