@@ -1,12 +1,11 @@
 use std::{
     ffi::CString,
     fmt,
+    os::unix::io::{BorrowedFd, OwnedFd},
     os::unix::{io::RawFd, net::UnixStream},
     sync::Arc,
 };
 
-use io_lifetimes::{BorrowedFd, OwnedFd};
-
 use crate::protocol::{Interface, Message, ObjectInfo};
 pub use crate::types::server::{Credentials, DisconnectReason, GlobalInfo, InitError, InvalidId};
 
@@ -32,11 +31,17 @@ pub trait ObjectData<D>: downcast_rs::DowncastSync {
         msg: Message<ObjectId, OwnedFd>,
     ) -> Option<Arc<dyn ObjectData<D>>>;
     /// Notification that the object has been destroyed and is no longer active
-    fn destroyed(&self, data: &mut D, client_id: ClientId, object_id: ObjectId);
+    fn destroyed(
+        self: Arc<Self>,
+        handle: &Handle,
+        data: &mut D,
+        client_id: ClientId,
+        object_id: ObjectId,
+    );
     /// Helper for forwarding a Debug implementation of your `ObjectData` type
     ///
     /// By default will just print `ObjectData { ... }`
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn debug(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("ObjectData").finish_non_exhaustive()
     }
@@ -45,7 +50,7 @@ pub trait ObjectData<D>: downcast_rs::DowncastSync {
 downcast_rs::impl_downcast!(sync ObjectData<D>);
 
 impl<D: 'static> std::fmt::Debug for dyn ObjectData<D> {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         self.debug(f)
     }
@@ -84,14 +89,14 @@ pub trait GlobalHandler<D>: downcast_rs::DowncastSync {
     /// Helper for forwarding a Debug implementation of your `GlobalHandler` type
     ///
     /// By default will just print `GlobalHandler { ... }`
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn debug(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("GlobalHandler").finish_non_exhaustive()
     }
 }
 
 impl<D: 'static> std::fmt::Debug for dyn GlobalHandler<D> {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.debug(f)
     }
@@ -108,14 +113,14 @@ pub trait ClientData: downcast_rs::DowncastSync {
     /// Helper for forwarding a Debug implementation of your `ClientData` type
     ///
     /// By default will just print `GlobalHandler { ... }`
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn debug(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
         f.debug_struct("ClientData").finish_non_exhaustive()
     }
 }
 
 impl std::fmt::Debug for dyn ClientData {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.debug(f)
     }
@@ -178,14 +183,14 @@ impl ObjectId {
 }
 
 impl fmt::Display for ObjectId {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.id.fmt(f)
     }
 }
 
 impl fmt::Debug for ObjectId {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.id.fmt(f)
     }
@@ -202,7 +207,7 @@ pub struct ClientId {
 }
 
 impl fmt::Debug for ClientId {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.id.fmt(f)
     }
@@ -215,7 +220,7 @@ pub struct GlobalId {
 }
 
 impl fmt::Debug for GlobalId {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         self.id.fmt(f)
     }
@@ -480,6 +485,13 @@ impl Handle {
     ) -> Result<Arc<dyn GlobalHandler<D>>, InvalidId> {
         self.handle.get_global_handler(id.id)
     }
+
+    /// Flushes pending events destined for a client.
+    ///
+    /// If no client is specified, all pending events are flushed to all clients.
+    pub fn flush(&mut self, client: Option<ClientId>) -> std::io::Result<()> {
+        self.handle.flush(client)
+    }
 }
 
 /// A backend object that represents the state of a wayland server.
@@ -567,7 +579,7 @@ impl<D> Backend<D> {
 pub(crate) struct DumbObjectData;
 
 impl<D> ObjectData<D> for DumbObjectData {
-    #[cfg_attr(coverage, no_coverage)]
+    #[cfg_attr(coverage, coverage(off))]
     fn request(
         self: Arc<Self>,
         _handle: &Handle,
@@ -578,6 +590,13 @@ impl<D> ObjectData<D> for DumbObjectData {
         unreachable!()
     }
 
-    #[cfg_attr(coverage, no_coverage)]
-    fn destroyed(&self, _: &mut D, _client_id: ClientId, _object_id: ObjectId) {}
+    #[cfg_attr(coverage, coverage(off))]
+    fn destroyed(
+        self: Arc<Self>,
+        _handle: &Handle,
+        _: &mut D,
+        _client_id: ClientId,
+        _object_id: ObjectId,
+    ) {
+    }
 }
